@@ -99,8 +99,15 @@ class Attention(nn.Module):
         self.compress_ratio = compress_ratio
         self.compress_fn = nn.Conv1d(inner_dim, inner_dim, compress_ratio, stride = compress_ratio, groups = heads) if compress_ratio > 1 else None
 
-    def forward(self, x, context = None, mask = None, context_mask = None):
-        device, h, has_context = x.device, self.heads, exists(context)
+    def forward(self, x, context = None, mask = None, context_mask = None, lead_dims = None):
+        device, orig_shape, h, has_context = x.device, x.shape, self.heads, exists(context)
+
+        if exists(lead_dims):
+            x = rearrange(x, 'b (m n) d -> (b m) n d', m = lead_dims[1])
+
+            if exists(mask):
+                mask = rearrange(mask, 'b (m n) -> (b m) n', m = lead_dims[1])
+
         context = default(context, x)
 
         q, k, v = (self.to_q(x), *self.to_kv(context).chunk(2, dim = -1))
@@ -134,8 +141,8 @@ class Attention(nn.Module):
         dots = einsum('b h i d, b h j d -> b h i j', q, k) * self.scale
 
         if exists(mask) or exists(context_mask):
-            mask = default(mask, lambda: torch.ones(i, device = device))
-            context_mask = default(context_mask, mask) if not has_context else default(context_mask, lambda: torch.ones(j, device = device))
+            mask = default(mask, lambda: torch.ones(1, i, device = device).bool())
+            context_mask = default(context_mask, mask) if not has_context else default(context_mask, lambda: torch.ones(1, j, device = device).bool())
             mask_value = -torch.finfo(dots.dtype).max
             mask = mask[:, None, :, None] * context_mask[:, None, None, :]
             dots.masked_fill_(~mask, mask_value)
@@ -145,7 +152,12 @@ class Attention(nn.Module):
 
         out = einsum('b h i j, b h j d -> b h i d', attn, v)
         out = rearrange(out, 'b h n d -> b n (h d)')
-        return self.to_out(out)
+        out = self.to_out(out)
+
+        if exists(lead_dims):
+            out = out.reshape(*orig_shape)
+
+        return out
 
 class SparseAttention(Attention):
     def __init__(
@@ -173,14 +185,17 @@ class SparseAttention(Attention):
         )
 
     def forward(self, x, mask = None):
-        device, h = x.device, self.heads
-        n = x.shape[1]
+        device, orig_shape, h = x.device, x.shape, self.heads
+
+        b, n, _ = x.shape
+        assert n <= self.seq_len, f'either the AA sequence length {n} or the total MSA length {n} exceeds the allowable sequence length {self.seq_len} for sparse attention, set by `max_seq_len`'
 
         remainder = x.shape[1] % self.block_size
 
         if remainder > 0:
             padding = self.block_size - remainder
             x = F.pad(x, (0, 0, 0, padding), value = 0)
+            mask = torch.ones(b, n, device = device).bool()
             mask = F.pad(mask, (0, padding), value = False)
 
         q, k, v = (self.to_q(x), *self.to_kv(x).chunk(2, dim = -1))
@@ -195,7 +210,9 @@ class SparseAttention(Attention):
         out = self.attn_fn(q, k, v, key_padding_mask = key_pad_mask)
         out = rearrange(out, 'b h n d -> b n (h d)')
         out = self.to_out(out)
-        return out[:, :n]
+        out = out[:, :n]
+
+        return out
 
 class AxialAttention(nn.Module):
     def __init__(
@@ -208,11 +225,12 @@ class AxialAttention(nn.Module):
         self.attn_width = attn_class(**kwargs)
         self.attn_height = attn_class(**kwargs)
 
-    def forward(self, x, context = None, mask = None, context_mask = None):
+    def forward(self, x, shape, context = None, mask = None, context_mask = None):
         b, n, d = x.shape
-        w = h = int(sqrt(n))
 
-        x = x.reshape(b, h, w, d)
+        x = x.view(shape)
+        h, w = shape[1:3]
+
         mask = mask.reshape(b, h, w) if exists(mask) else None
 
         w_mask = h_mask = w_context = h_context = w_context_mask = h_context_mask = None
@@ -246,15 +264,24 @@ class SequentialSequence(nn.Module):
         super().__init__()
         self.blocks = blocks
 
-    def forward(self, x, m, mask = None,  msa_mask = None, **kwargs):
+    def forward(
+        self,
+        x,
+        m,
+        seq_shape,
+        msa_shape,
+        mask = None,
+        msa_mask = None,
+        **kwargs
+    ):
         for ((attn, ff, msa_attn), (cross_attn, msa_ff, msa_cross_attn)) in zip(*[iter(self.blocks)] * 2):
 
             # self attention
 
-            x = attn(x, mask = mask) + x
+            x = attn(x, seq_shape, mask = mask) + x
 
             if exists(m):
-                m = msa_attn(m, mask = msa_mask) + m
+                m = msa_attn(m, msa_shape, mask = msa_mask) + m
 
                 # cross attention
 
@@ -286,7 +313,7 @@ class Alphafold2(nn.Module):
         ff_dropout = 0.,
         reversible = False,
         sparse_self_attn = False,
-        cross_attn_compress_ratio = 1,
+        cross_attn_compress_ratio = 1
     ):
         super().__init__()
         layers_sparse_attn = cast_tuple(sparse_self_attn, depth)
@@ -314,12 +341,10 @@ class Alphafold2(nn.Module):
 
             # self attention, for both main sequence and msa
 
-            attn_class = SparseAttention if layer_sparse_attn else Attention
-
             layers.append(nn.ModuleList([
                 prenorm(AxialAttention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, sparse_attn = sparse_self_attn)),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
-                prenorm(attn_class(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
+                prenorm(AxialAttention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
             ]))
 
@@ -368,6 +393,7 @@ class Alphafold2(nn.Module):
         x = rearrange(ax1, 'b i d -> b i () d') + rearrange(ax2, 'b j d-> b () j d') # create pair-wise residue embeds
         x_mask = rearrange(mask, 'b i -> b i ()') + rearrange(mask, 'b j -> b () j') if exists(mask) else None
 
+        seq_shape = x.shape
         x = rearrange(x, 'b i j d -> b (i j) d')
         x_mask = rearrange(x_mask, 'b i j -> b (i j)') if exists(mask) else None
 
@@ -375,12 +401,11 @@ class Alphafold2(nn.Module):
 
         m = None
         if exists(msa):
-            msa_shape = msa.shape
-
             m = self.token_emb(msa)
-            m += self.msa_pos_emb(torch.arange(msa_shape[-1], device = device))[None, None, ...]
-            m += self.msa_num_pos_emb(torch.arange(msa_shape[1], device = device))[None, :, None, :]
+            m += self.msa_pos_emb(torch.arange(msa.shape[-1], device = device))[None, None, ...]
+            m += self.msa_num_pos_emb(torch.arange(msa.shape[1], device = device))[None, :, None, :]
 
+            msa_shape = m.shape
             m = rearrange(m, 'b m n d -> b (m n) d')
 
         elif exists(embedds):
@@ -393,7 +418,14 @@ class Alphafold2(nn.Module):
 
         # trunk
 
-        x, m = self.net(x, m, mask = x_mask, msa_mask = msa_mask)
+        x, m = self.net(
+            x,
+            m,
+            seq_shape,
+            msa_shape,
+            mask = x_mask,
+            msa_mask = msa_mask
+        )
 
         # structural refinement
 
